# Create intensified image by registering a set of frames and summing them

import numpy as np
import astropy.io.fits as pyfits
from time import gmtime, strftime  # for utc
import cv2
# from skimage.filters import unsharp_mask


def asinhScale(img, limcut = 0):  # img needs to be float32 type
    # Try an asinh transformation to handle a wide dynamic range
    limg = np.arcsinh(img)
    limg = limg / limg.max()
    limg[limg < limcut] = 0.
    return limg

def frameStacker(pr, progress_bar, event_process,
                 first_frame, last_frame, timestamp_trim_top, timestamp_trim_bottom,
                 fitsReader, serReader, advReader,
                 avi_location, out_dir_path, bkg_threshold, hot_pixel_erase, delta_x, delta_y, shift_dict):
    # fitsReader is self.getFitsFrame()
    # serReader is self.getSerFrame()
    # advReader is self.getAdvFrame()
    # pr is self.showMsg() provided by the caller
    # hot_pixel_erase is self.applyHotPixelErasureToImg
    # progress_bar is a reference to the caller's progress bar item so
    # that can update it to show progess

    cap = None

    def read_fits_frame(frame_to_read, trim_top=0, trim_bottom=0):
        try:
            frame_local = fitsReader(frame_to_read)
            if frame_local is None:
                pr(f'Problem reading FITS file: frame == None returned')
                return None

            # if not frame_to_read == first_frame:
            #     cleaned = hot_pixel_erase(frame_local)
            # else:
            #     cleaned = frame_local

            cleaned = hot_pixel_erase(frame_local)

            image = cleaned[:, :].astype('float32')

            if trim_bottom:
                image = image[0:-trim_bottom, :]

            if trim_top:
                image = image[trim_top:, :]

        except Exception as e:
            pr(f'Problem reading FITS file: {e}')
            return None

        return image

    def read_ser_frame(frame_to_read, trim_top=0, trim_bottom=0):
        try:
            frame_local = serReader(frame_to_read)
            if frame_local is None:
                pr(f'Problem reading SER file: frame == None returned')
                return None

            # if not frame_to_read == first_frame:
            #     cleaned = hot_pixel_erase(frame_local)
            # else:
            #     cleaned = frame_local

            cleaned = hot_pixel_erase(frame_local)

            image = cleaned[:, :].astype('float32')

            if trim_bottom:
                image = image[0:-trim_bottom, :]

            if trim_top:
                image = image[trim_top:, :]

        except Exception as e:
            pr(f'Problem reading SER file: {e}')
            return None

        return image

    def read_adv_frame(frame_to_read, trim_top=0, trim_bottom=0):
        try:
            frame_local = advReader(frame_to_read)
            if frame_local is None:
                pr(f'Problem reading ADV file: frame == None returned')
                return None

            # if not frame_to_read == first_frame:
            #     cleaned = hot_pixel_erase(frame_local)
            # else:
            #     cleaned = frame_local

            cleaned = hot_pixel_erase(frame_local)

            image = cleaned[:, :].astype('float32')

            if trim_bottom:
                image = image[0:-trim_bottom, :]

            if trim_top:
                image = image[trim_top:, :]

        except Exception as e:
            pr(f'Problem reading ADV file: {e}')
            return None

        return image

    def read_avi_frame(frame_to_read, trim_top=0, trim_bottom=0):
        # roi = [xleft, xright, ytop, ybottom]
        status = None
        image = None
        try:
            cap.set(cv2.CAP_PROP_POS_FRAMES, frame_to_read)
            status, frame_local = cap.read()

            if len(frame_local.shape) == 3:
                frame_local = cv2.cvtColor(frame_local, cv2.COLOR_BGR2GRAY)

            # if not frame_to_read == first_frame:
            #     cleaned = hot_pixel_erase(frame_local)
            # else:
            #     cleaned = frame_local

            cleaned = hot_pixel_erase(frame_local)

            image = cleaned[:, :].astype('float32')

            if trim_bottom:
                image = image[0:-trim_bottom, :]

            if trim_top:
                image = image[trim_top:, :]

        except Exception as e:
            pr(f'Problem reading avi file: {e}')

        if status:
            return image
        else:
            return None

    def openAviReader(avi_location_param):
        pr(f'Trying to open: {avi_location_param}')
        cap_local = cv2.VideoCapture(avi_location_param)
        if not cap_local.isOpened():
            pr(f'  {avi_location_param} could not be opened!')
            return None
        else:
            pr(f'...file opened just fine.')
            return cap_local

    if shift_dict:
        xc = shift_dict['x']
        yc = shift_dict['y']
        frame = shift_dict['frame']
    else:
        xc = []
        yc = []
        frame = []

    if shift_dict:
        err = False
        if not first_frame == frame[0]:
            pr(f'ERROR(not equal): frame[0]: {frame[0]}  first_frame: {first_frame}')
            err = True
        if not last_frame == frame[-1]:
            pr(f'ERROR(not equal): frame[-1]: {frame[-1]}  last_frame: {last_frame}')
            err = True
        if err:
            return

    if fitsReader is None and serReader is None and advReader is None:
        cap = openAviReader(avi_location_param=avi_location)
        if cap is None:
            return

    next_frame = first_frame + 1

    # Read reference frame image without trimming the timestamp portion off
    # so that we can save the timestamp for later replacement.
    if fitsReader:
        inimage = read_fits_frame(first_frame)
    elif serReader:
        inimage = read_ser_frame(first_frame)
    elif advReader:
        inimage = read_adv_frame(first_frame)
    else:
        inimage = read_avi_frame(first_frame)

    height, width = inimage.shape
    pr(f'image shape: {width} x {height}')

    if timestamp_trim_top > 0:
        # If redact is from the top, this is assumed to be a FITS or SER file for
        # which there is no timestamp to be preserved, just a few 'corrupted' lines at the top.
        timestamp_image_top = inimage[0:timestamp_trim_top,:]
    else:
        timestamp_image_top = None

    if timestamp_trim_bottom > 0:
        timestamp_image_bottom = inimage[-timestamp_trim_bottom:,:]
    else:
        timestamp_image_bottom = None

    # Re-read the reference frame with the timestamp trimmed off and use it
    # to initialize the stack sum
    if fitsReader:
        inimage = read_fits_frame(first_frame, trim_top=timestamp_trim_top, trim_bottom=timestamp_trim_bottom)
    elif serReader:
        inimage = read_ser_frame(first_frame, trim_top=timestamp_trim_top, trim_bottom=timestamp_trim_bottom)
    elif advReader:
        inimage = read_adv_frame(first_frame, trim_top=timestamp_trim_top, trim_bottom=timestamp_trim_bottom)
    else:
        inimage = read_avi_frame(first_frame, trim_top=timestamp_trim_top, trim_bottom=timestamp_trim_bottom)

    image_sum = inimage[:,:]

    # frames_skipped = 0

    # g1 is our reference image transform
    # g1 = np.fft.fftshift(np.fft.fft2(inimage))
    # if not shift_dict:
        # ret, th_inimage = cv2.threshold(inimage, bkg_threshold, 0, cv2.THRESH_TOZERO)
        # g1 = np.fft.fftshift(np.fft.fft2(th_inimage))

    k = 0
    while next_frame <= last_frame:
        if fitsReader:
            inimage = read_fits_frame(next_frame, trim_top=timestamp_trim_top, trim_bottom=timestamp_trim_bottom)
        elif serReader:
            inimage = read_ser_frame(next_frame, trim_top=timestamp_trim_top, trim_bottom=timestamp_trim_bottom)
        elif advReader:
            inimage = read_adv_frame(next_frame, trim_top=timestamp_trim_top, trim_bottom=timestamp_trim_bottom)
        else:
            inimage = read_avi_frame(next_frame, trim_top=timestamp_trim_top, trim_bottom=timestamp_trim_bottom)

        delta_frame = next_frame - first_frame

        next_frame += 1
        k += 1

        # Calculate progress [1..100]
        fraction_done = (next_frame - first_frame) / (last_frame - first_frame)
        progress_bar.setValue(int(fraction_done * 100))
        event_process()

        rows_to_roll_to_center = None
        cols_to_roll_to_center = None

        if delta_x is None:
            # g2 = np.fft.fftshift(np.fft.fft2(inimage))
            g1 = None
            if not shift_dict:
                ret, th_inimage = cv2.threshold(inimage, bkg_threshold, 0, cv2.THRESH_TOZERO)
                g2 = np.fft.fftshift(np.fft.fft2(th_inimage))

                g2conj = g2.conj()

                R = g1 * g2conj / abs(g1 * g2conj)

                r = np.fft.fftshift(np.fft.ifft2(R))

                mag_r = abs(r * r.conj())  # mag_r is a matrix of positive reals (not complex)

                # mag_r_max = mag_r.max()

                max_row, max_col = np.unravel_index(mag_r.argmax(), mag_r.shape)

                # print(mag_r.shape, max_row, max_col)
                rows_to_roll_to_center = max_row - int(mag_r.shape[0] / 2)
                cols_to_roll_to_center = max_col - int(mag_r.shape[1] / 2)
            else:
                rows_to_roll_to_center = yc[0] - yc[k]
                cols_to_roll_to_center = xc[0] - xc[k]

        if not delta_x is None:
            rows_to_roll_to_center = round(-delta_y * delta_frame)
            cols_to_roll_to_center = round(-delta_x * delta_frame)


        pr(f'row-shift:{rows_to_roll_to_center:4d}  col-shift:{cols_to_roll_to_center:4d}  frame: {next_frame-1}',
            blankLine=False)

        # Center the image
        inimage = np.roll(inimage, rows_to_roll_to_center, axis=0)
        inimage = np.roll(inimage, cols_to_roll_to_center, axis=1)

        # ... and add it to the image sum
        image_sum += inimage

    progress_bar.setValue(0)

    # Instead of sharpening, we just divide by the number of frames that were summed
    normed_image = image_sum / (last_frame - first_frame + 1)

    fn = f'/enhanced-image-{first_frame}.fit'

    outfile = out_dir_path + fn

    if not timestamp_image_bottom is None:
        normed_image = np.append(normed_image, timestamp_image_bottom, axis=0)
    if not timestamp_image_top is None:
        normed_image = np.append(timestamp_image_top, normed_image, axis=0)
    outlist = pyfits.PrimaryHDU(normed_image)

    # Provide a new date stamp
    file_time = strftime("%Y-%m-%d %H:%M:%S", gmtime())

    # Compose the FITS header
    outhdr = outlist.header

    # Add the REQUIRED elements in the REQUIRED order
    outhdr['SIMPLE'] = True
    outhdr['NAXIS']  = 2
    outhdr['NAXIS1'] = width
    outhdr['NAXIS2'] = height
    # End of required elements

    outhdr['DATE'] = file_time
    outhdr['FILE'] = avi_location
    outhdr['COMMENT'] = f'{last_frame - first_frame + 1} frames were stacked'
    outhdr['COMMENT'] = f'Initial frame number: {first_frame}'
    outhdr['COMMENT'] = f'Final frame number: {last_frame}'

    # Write the fits file
    outlist.writeto(outfile, overwrite = True)

def hotPixelStack(pr, progress_bar, event_process,
                 first_frame, last_frame, timestamp_trim_top, timestamp_trim_bottom,
                 fitsReader, serReader,
                 avi_location, out_dir_path, bkg_threshold):

    _ = out_dir_path  # unused parameter
    _ = bkg_threshold # unused parameter

    # fitsReader is self.getFitsFrame()
    # serReader is self.getSerFrame()
    # pr is self.showMsg() provided by the caller
    # progress_bar is a reference to the caller's progress bar item so
    # that can update it to show progess

    pr(f'hot pixel stacker called.')
    # return

    cap = None

    def read_fits_frame(frame_to_read, trim_top=0, trim_bottom=0):
        try:
            frame = fitsReader(frame_to_read)
            if frame is None:
                pr(f'Problem reading FITS file: frame == None returned')
                return None

            image = frame[:, :].astype('float32')

            if trim_bottom:
                image = image[0:-trim_bottom, :]

            if trim_top:
                image = image[trim_top:, :]
            # if trim > 0:
            #     image = frame[0:-trim, :].astype('float32')
            # else:
            #     image = frame[-trim:, :].astype('float32')

        except Exception as e:
            pr(f'Problem reading FITS file: {e}')
            return None

        return image

    def read_ser_frame(frame_to_read, trim_top=0, trim_bottom=0):
        try:
            frame = serReader(frame_to_read)
            if frame is None:
                pr(f'Problem reading SER file: frame == None returned')
                return None

            image = frame[:, :].astype('float32')

            if trim_bottom:
                image = image[0:-trim_bottom, :]

            if trim_top:
                image = image[trim_top:, :]
            # if trim is None or trim == 0:
            #     image = frame[:, :].astype('float32')
            # else:
            #     if trim > 0:
            #         image = frame[0:-trim, :].astype('float32')
            #     else:
            #         image = frame[-trim:, :].astype('float32')

        except Exception as e:
            pr(f'Problem reading SER file: {e}')
            return None

        return image

    def read_avi_frame(frame_to_read, trim_top=0, trim_bottom=0):
        status = None
        image = None
        try:
            cap.set(cv2.CAP_PROP_POS_FRAMES, frame_to_read)
            status, frame = cap.read()

            if len(frame.shape) == 3:
                frame = cv2.cvtColor(frame, cv2.COLOR_BGR2GRAY)

            image = frame[:, :].astype('float32')

            if trim_bottom:
                image = image[0:-trim_bottom, :]

            if trim_top:
                image = image[trim_top:, :]

            # if trim is None or trim == 0:
            #     image = frame[:, :, 0].astype('float32')
            # else:
            #     image = frame[0:-trim, :, 0].astype('float32')
            # plt.imshow(asinhScale(image), cmap='gray')
        except Exception as e:
            pr(f'Problem reading avi file: {e}')

        if status:
            return image
        else:
            return None

    def openAviReader(avi_location_param):
        pr(f'Trying to open: {avi_location_param}')
        cap_local = cv2.VideoCapture(avi_location_param)
        if not cap_local.isOpened():
            pr(f'  {avi_location_param} could not be opened!')
            return None
        else:
            pr(f'...file opened just fine.')
            return cap_local

    if fitsReader is None and serReader is None:
        cap = openAviReader(avi_location_param=avi_location)
        if cap is None:
            return

    next_frame = first_frame + 1

    # Read reference frame image without trimming the timestamp portion off
    # so that we can save the timestamp for later replacement.
    if fitsReader:
        read_fits_frame(first_frame)
    elif serReader:
        read_ser_frame(first_frame)
    else:
        read_avi_frame(first_frame)

    # if timestamp_trim_top > 0:
        # If redact is from the top, this is assumed to be a FITS or SER file for
        # which there is no timestamp to be preserved, just a few 'corrupted' lines at the top.
        # timestamp_image_top = inimage[0:timestamp_trim_top,:]
        # timestamp_image = np.zeros_like(timestamp_junk)
    # else:
        # timestamp_image_top = None

    # if timestamp_trim_bottom > 0:
    #     timestamp_image_bottom = inimage[-timestamp_trim_bottom:,:]
    # else:
    #     timestamp_image_bottom = None

    # Re-read the reference frame with the timestamp trimmed off and use it
    # to initialize the stack sum
    if fitsReader:
        inimage = read_fits_frame(first_frame, trim_top=timestamp_trim_top, trim_bottom=timestamp_trim_bottom)
    elif serReader:
        inimage = read_ser_frame(first_frame, trim_top=timestamp_trim_top, trim_bottom=timestamp_trim_bottom)
    else:
        inimage = read_avi_frame(first_frame, trim_top=timestamp_trim_top, trim_bottom=timestamp_trim_bottom)

    image_sum = inimage[:,:]

    # ret, th_inimage = cv2.threshold(inimage, bkg_threshold, 0, cv2.THRESH_TOZERO)

    while next_frame <= last_frame:
        if fitsReader:
            inimage = read_fits_frame(next_frame, trim_top=timestamp_trim_top, trim_bottom=timestamp_trim_bottom)
        elif serReader:
            inimage = read_ser_frame(next_frame, trim_top=timestamp_trim_top, trim_bottom=timestamp_trim_bottom)
        else:
            inimage = read_avi_frame(next_frame, trim_top=timestamp_trim_top, trim_bottom=timestamp_trim_bottom)

        next_frame += 1

        # Calculate progress [1..100]
        fraction_done = (next_frame - first_frame) / (last_frame - first_frame)
        progress_bar.setValue(fraction_done * 100)
        event_process()

        # ret, th_inimage = cv2.threshold(inimage, bkg_threshold, 0, cv2.THRESH_TOZERO)
        # ... and add it to the image sum
        image_sum += inimage

    progress_bar.setValue(0)

    avg_image = image_sum / (last_frame - first_frame + 1)

    # if not timestamp_image_bottom is None:
    #     unredacted = np.append(unredacted, asinhScale(timestamp_image_bottom), axis=0)
    # if not timestamp_image_top is None:
    #     unredacted = np.append(asinhScale(timestamp_image_top), unredacted, axis=0)

    # outfile = out_dir_path + r'/enhanced-image.fit'

    # min_pixel = unredacted.min()
    # unredacted = unredacted - min_pixel
    # max_pixel = unredacted.max()
    # unredacted = unredacted * 255 / max_pixel

    return avg_image.astype('uint8')

def find_outlier_pixels(data,tolerance=3,worry_about_edges=True):
    #This function finds the hot or dead pixels in a 2D dataset.
    #tolerance is the number of standard deviations used to cutoff the hot pixels
    #If you want to ignore the edges and greatly speed up the code, then set
    #worry_about_edges to False.
    #
    #The function returns a list of hot pixels and also an image with with hot pixels removed
    _ = tolerance # unused parameter

    from scipy.ndimage import median_filter
    blurred = median_filter(data, size=3)
    difference = data - blurred
    threshold = 10 * np.std(difference)

    #find the hot pixels, but ignore the edges
    hot_pixels = np.nonzero((np.abs(difference[1:-1,1:-1])>threshold) )
    hot_pixels = np.array(hot_pixels) + 1 #because we ignored the first row and first column

    fixed_image = np.copy(data) #This is the image with the hot pixels removed
    for y,x in zip(hot_pixels[0],hot_pixels[1]):
        fixed_image[y,x]=blurred[y,x]

    if worry_about_edges:
        height,width = np.shape(data)

        ###Now get the pixels on the edges (but not the corners)###

        #left and right sides
        for index in range(1,height-1):
            #left side:
            med  = np.median(data[index-1:index+2,0:2])
            diff = np.abs(data[index,0] - med)
            if diff>threshold:
                hot_pixels = np.hstack(( hot_pixels, [[index],[0]]  ))
                fixed_image[index,0] = med

            #right side:
            med  = np.median(data[index-1:index+2,-2:])
            diff = np.abs(data[index,-1] - med)
            if diff>threshold:
                hot_pixels = np.hstack(( hot_pixels, [[index],[width-1]]  ))
                fixed_image[index,-1] = med

        #Then the top and bottom
        for index in range(1,width-1):
            #bottom:
            med  = np.median(data[0:2,index-1:index+2])
            diff = np.abs(data[0,index] - med)
            if diff>threshold:
                hot_pixels = np.hstack(( hot_pixels, [[0],[index]]  ))
                fixed_image[0,index] = med

            #top:
            med  = np.median(data[-2:,index-1:index+2])
            diff = np.abs(data[-1,index] - med)
            if diff>threshold:
                hot_pixels = np.hstack(( hot_pixels, [[height-1],[index]]  ))
                fixed_image[-1,index] = med

        ###Then the corners###

        #bottom left
        med  = np.median(data[0:2,0:2])
        diff = np.abs(data[0,0] - med)
        if diff>threshold:
            hot_pixels = np.hstack(( hot_pixels, [[0],[0]]  ))
            fixed_image[0,0] = med

        #bottom right
        med  = np.median(data[0:2,-2:])
        diff = np.abs(data[0,-1] - med)
        if diff>threshold:
            hot_pixels = np.hstack(( hot_pixels, [[0],[width-1]]  ))
            fixed_image[0,-1] = med

        #top left
        med  = np.median(data[-2:,0:2])
        diff = np.abs(data[-1,0] - med)
        if diff>threshold:
            hot_pixels = np.hstack(( hot_pixels, [[height-1],[0]]  ))
            fixed_image[-1,0] = med

        #top right
        med  = np.median(data[-2:,-2:])
        diff = np.abs(data[-1,-1] - med)
        if diff>threshold:
            hot_pixels = np.hstack(( hot_pixels, [[height-1],[width-1]]  ))
            fixed_image[-1,-1] = med

    return hot_pixels, fixed_image.astype('uint8')


from pymovie import apertureEditDialog
from PyQt5.QtWidgets import QDialog
from PyQt5 import QtGui
from PyQt5 import QtCore
from PyQt5.QtWidgets import QTableWidgetItem, QMenu, QAction
import numpy as np


class EditApertureDialog(QDialog, apertureEditDialog.Ui_Dialog):
    def __init__(self, messager, saver, dictList, appSize, threshSpinner, imageUpdate, setThumbnails):
        super(EditApertureDialog, self).__init__()
        self.setupUi(self)
        self.msgRoutine = messager
        self.settingsSaver = saver
        self.dictList = dictList
        self.fillApertureTable()
        self.appSize = appSize
        self.threshSpinner = threshSpinner
        self.imageUpdate = imageUpdate
        self.tableWidget.cellClicked.connect(self.cellClicked)
        self.tableWidget.cellChanged.connect(self.cellClicked)
        self.tableWidget.cellActivated.connect(self.cellClicked)
        self.tableWidget.itemSelectionChanged.connect(self.selectionChange)
        self.setThumbnails = setThumbnails
        self.ignoreCellClick = False
        self.col = None
        self.row = None
        self.menu = None

    def selectionChange(self):
        row = self.tableWidget.currentRow()
        col = self.tableWidget.currentColumn()
        self.cellClicked(row, col)

    def cellClicked(self, row, column):
        if self.ignoreCellClick:
            self.ignoreCellClick = False
            return
        # self.msgRoutine( f'row {row} column {column} was clicked')
        aperture = self.dictList[row]['appRef']
        self.writeTable()
        showDefaultMaskInThumbnailTwo = column == 3
        self.setThumbnails(aperture, showDefaultMaskInThumbnailTwo)
        # The xy position may have changed because of 'snap' when threshold is changed.
        self.ignoreCellClick = True
        xc, yc = aperture.getCenter()
        item = QTableWidgetItem(str(f'({xc},{yc})'))
        self.tableWidget.setItem(row, 1, item)

    def writeTable(self):
        self.updateAperturesFromTable()

    def fillApertureTable(self):
        for rowDict in self.dictList:
            numRows = self.tableWidget.rowCount()
            self.tableWidget.insertRow(numRows)

            item = QTableWidgetItem(str(rowDict['name']))
            self.tableWidget.setItem(numRows, 0, item)

            item = QTableWidgetItem(str(rowDict['xy']))
            self.tableWidget.setItem(numRows, 1, item)

            item = QTableWidgetItem(str(rowDict['threshDelta']))
            self.tableWidget.setItem(numRows, 2, item)

            item = QTableWidgetItem(str(rowDict['defMskRadius']))
            self.tableWidget.setItem(numRows, 3, item)

            color_str = str(rowDict['color'])
            if color_str.startswith('red'):
                item = QTableWidgetItem('red (standard)')
            elif color_str.startswith('green'):
                item = QTableWidgetItem('green (connect to threshold spinner)')
            elif color_str.startswith('yellow'):
                item = QTableWidgetItem('yellow (tracking aperture)')
            elif color_str.startswith('white'):
                item = QTableWidgetItem('white (special flash tag aperture')
            # item = QTableWidgetItem(str(rowDict['color']))
            item.setFlags(item.flags() ^ QtCore.Qt.ItemIsEditable)
            self.tableWidget.setItem(numRows, 4, item)

            item = QTableWidgetItem(str(rowDict['joggable']))
            item.setFlags(item.flags() ^ QtCore.Qt.ItemIsEditable)
            self.tableWidget.setItem(numRows, 5, item)

            item = QTableWidgetItem(str(rowDict['autoTextOut']))
            item.setFlags(item.flags() ^ QtCore.Qt.ItemIsEditable)
            self.tableWidget.setItem(numRows, 6, item)

            item = QTableWidgetItem(str(rowDict['thumbnailSource']))
            item.setFlags(item.flags() ^ QtCore.Qt.ItemIsEditable)
            self.tableWidget.setItem(numRows, 7, item)

            item = QTableWidgetItem(str(rowDict['outputOrder']))
            self.tableWidget.setItem(numRows, 8, item)

    def createDefaultMask(self, radius):
        mask = np.zeros((self.appSize, self.appSize), 'int16')
        maskPixelCount = 0
        roi_center = int(self.appSize / 2)
        c = roi_center
        r = int(np.ceil(radius))
        if r > c - 1:
            r = c - 1
            radius = r
        for i in range(c - r - 1, c + r + 2):
            for j in range(c - r - 1, c + r + 2):
                if (i - c) ** 2 + (j - c) ** 2 <= radius ** 2:
                    maskPixelCount += 1
                    mask[i, j] = 1
        return mask, maskPixelCount, radius

    def closeEvent(self, event):
        self.settingsSaver.setValue('appEditDialogSize', self.size())
        self.settingsSaver.setValue('appEditDialogPos', self.pos())
        self.updateAperturesFromTable()

    def parseXY(self, xyText):
        # noinspection PyBroadException
        try:
            parts = xyText.split(",")
            xc = int(parts[0].strip().strip("(").strip())
            yc = int(parts[1].strip().strip(")").strip())
        except:
            self.msgRoutine(f'"{xyText}" is an invalid format for x,y')
            return None, None
        return xc, yc

    def updateAperturesFromTable(self):

        # Now comes the hard work of validating and using the tableWidget data
        # to update the aperture properties

        for row in range(self.tableWidget.rowCount()):

            aperture = self.dictList[row]['appRef']

            xyText = self.tableWidget.item(row, 1).text()
            xc, yc = self.parseXY(xyText=xyText)
            if xc is not None:
                aperture.setCenter(xc, yc)
                self.imageUpdate()

            aperture.name = self.tableWidget.item(row, 0).text()

            # Enforce that thresh is a positive integer
            text = self.tableWidget.item(row, 2).text()
            try:
                thresh = int(text)
            except ValueError:
                self.msgRoutine(f'In {aperture.name}(thresh): {text} is not a valid integer')
                return
            if thresh < 0:
                self.msgRoutine(f'In {aperture.name}(thresh): {text} is not a positive integer')
                return
            aperture.thresh = thresh

            # Enforce that def mask radius has value between 2.0 and 24.0
            text = self.tableWidget.item(row, 3).text()
            try:
                radius = float(text)
            except ValueError:
                self.msgRoutine(f'In {aperture.name}(def mask radius): {text} is not a valid float')
                return
            if radius < 2.0:
                self.msgRoutine(f'In {aperture.name}(def mask radius): {text} cannot be less than 2.0')
                return
            aperture.default_mask_radius = radius

            aperture.defaultMask, aperture.defaultMaskPixelCount, aperture.default_mask_radius = \
                self.createDefaultMask(radius)

            aperture.color = self.tableWidget.item(row, 4).text()
            if aperture.color.startswith('green'):
                aperture.setGreen()
                self.updateSpinnersFromRow(row)
            elif aperture.color.startswith('red'):
                aperture.setRed()
            elif aperture.color.startswith('white'):
                aperture.setWhite()
            elif aperture.color.startswith('yellow'):
                aperture.setYellowNoCheck()

            text = self.tableWidget.item(row, 5).text()
            if text == 'True':
                aperture.jogging_enabled = True
            else:
                aperture.jogging_enabled = False

            text = self.tableWidget.item(row, 6).text()
            if text == 'True':
                aperture.auto_display = True
            else:
                aperture.auto_display = False

            text = self.tableWidget.item(row, 7).text()
            if text == 'True':
                aperture.thumbnail_source = True
            else:
                aperture.thumbnail_source = False

            # Enforce that output order is a positive integer
            text = self.tableWidget.item(row, 8).text()
            try:
                order = int(text)
            except ValueError:
                self.msgRoutine(f'In {aperture.name}(order): {text} is not a valid integer')
                return
            if order < 0:
                self.msgRoutine(f'In {aperture.name}(order): {text} is not a positive integer')
                return
            aperture.order_number = order

    def contextMenuEvent(self, event):
        # self.msgRoutine("Got a right-click event")
        self.col = self.tableWidget.currentColumn()
        self.row = self.tableWidget.currentRow()
        items = self.tableWidget.selectedItems()
        if not items:
            self.msgRoutine('Nothing selected')
            return
        # self.msgRoutine(f'row: {self.row}  column: {self.col} items: {items[0]}')

        if 5 <= self.col <= 7:
            self.menu = QMenu()
            # self.menu = QtGui.QMenu()
            doTrue = QAction("Set True", self)
            doTrue.triggered.connect(self.setTrue)
            self.menu.addAction(doTrue)

            doFalse = QAction("Set False", self)
            doFalse.triggered.connect(self.setFalse)
            self.menu.addAction(doFalse)

            self.menu.popup(QtGui.QCursor.pos())
        elif self.col == 4:
            self.menu = QMenu()
            # self.menu = QtGui.QMenu()

            setRed = QAction("Set red (standard)", self)
            setRed.triggered.connect(self.setRed)
            self.menu.addAction(setRed)

            setGreen = QAction("Set green (connect to threshold spinner)", self)
            setGreen.triggered.connect(self.setGreen)
            self.menu.addAction(setGreen)

            setYellow = QAction("Set yellow (tracking aperture)", self)
            setYellow.triggered.connect(self.setYellow)
            self.menu.addAction(setYellow)

            setWhite = QAction("Set white (special flash tag aperture)", self)
            setWhite.triggered.connect(self.setWhite)
            self.menu.addAction(setWhite)

            self.menu.popup(QtGui.QCursor.pos())

    def setTrue(self):
        if self.col == 7:
            # We enforce the condition that only one aperture
            # can have True for this property
            for row in range(self.tableWidget.rowCount()):
                item = QTableWidgetItem('False')
                item.setFlags(item.flags() ^ QtCore.Qt.ItemIsEditable)
                self.tableWidget.setItem(row, self.col, item)

        item = QTableWidgetItem('True')
        item.setFlags(item.flags() ^ QtCore.Qt.ItemIsEditable)
        self.tableWidget.setItem(self.row, self.col, item)

    def setFalse(self):
        item = QTableWidgetItem('False')
        item.setFlags(item.flags() ^ QtCore.Qt.ItemIsEditable)
        self.tableWidget.setItem(self.row, self.col, item)

    def setRed(self):
        item = QTableWidgetItem('red (standard)')
        item.setFlags(item.flags() ^ QtCore.Qt.ItemIsEditable)
        self.tableWidget.setItem(self.row, self.col, item)

    def setGreen(self):
        for row in range(self.tableWidget.rowCount()):
            # To automatically enforce the 'only one green' policy, turn any existing
            # 'green' to 'red'
            if self.tableWidget.item(row, self.col).text().startswith('green'):
                item = QTableWidgetItem('red (standard)')
                item.setFlags(item.flags() ^ QtCore.Qt.ItemIsEditable)
                self.tableWidget.setItem(row, self.col, item)
        item = QTableWidgetItem('green (connect to threshold spinner)')
        item.setFlags(item.flags() ^ QtCore.Qt.ItemIsEditable)
        self.tableWidget.setItem(self.row, self.col, item)

        self.updateSpinnersFromRow(self.row)

    def updateSpinnersFromRow(self, row):

        try:
            thresh = int(self.tableWidget.item(row, 2).text())
        except ValueError:
            thresh = 0
        if thresh < 0:
            thresh = 0

        self.threshSpinner.setValue(thresh)

    def setYellow(self):
        numYellow = 0
        for row in range(self.tableWidget.rowCount()):
            if self.tableWidget.item(row, self.col).text().startswith('yellow'):
                numYellow += 1

        if numYellow == 2:
            self.msgRoutine(f'!!! There can only be a max of two yellow apertures !!!')
            return

        item = QTableWidgetItem('yellow (tracking aperture)')
        item.setFlags(item.flags() ^ QtCore.Qt.ItemIsEditable)
        self.tableWidget.setItem(self.row, self.col, item)

    def setWhite(self):
        item = QTableWidgetItem('white (special flash tag aperture)')
        item.setFlags(item.flags() ^ QtCore.Qt.ItemIsEditable)
        self.tableWidget.setItem(self.row, self.col, item)



import numpy as np
import cv2
import matplotlib.pyplot as plt

KiwiJumpAtMinuteChange = 11

# noinspection PyTypeChecker
def setup_for_iota_720_safe_mode3():
    # Do initializations needed for IOTA VTI timestamp extraction
    # Parameters for IOTA VTI timestamp characters when in safe mode

    # Define xy coordinates of upper field character box corners
    xcU = [72, 96, 146, 170, 220, 244, 294, 318, 343, 367, 417, 441, 465, 490]
    ycU = [199] * 14

    # Define xy coordinates of lower field character box corners
    xcL = [72, 96, 146, 170, 220, 244, 294, 318, 343, 367, 417, 441, 465, 490]
    ycL = [199] * 14

    upper_field_boxes = [None] * len(xcL)
    lower_field_boxes = [None] * len(xcL)

    # Turn box corners into full box coordinate tuples
    for i in range(len(xcL)):
        upper_field_boxes[i] = (xcU[i], xcU[i] + 21, ycU[i], ycU[i] + 14)
        lower_field_boxes[i] = (xcL[i], xcL[i] + 21, ycL[i], ycL[i] + 14)

    return upper_field_boxes, lower_field_boxes


# noinspection PyTypeChecker,PyTypeChecker
def setup_for_iota_640_safe_mode3():
    # Do initializations needed for IOTA VTI timestamp extraction
    # Parameters for IOTA VTI timestamp characters when in safe mode

    # Define xy coordinates of upper field character box corners
    xcU = [64, 87, 129, 151, 194, 217, 260, 282, 303, 325, 369, 391, 413, 435]
    ycU = [199] * 14

    # Define xy coordinates of lower field character box corners
    xcL = [64, 87, 129, 151, 194, 217, 260, 282, 303, 325, 369, 391, 413, 435]
    ycL = [199] * 14

    upper_field_boxes = [None] * len(xcL)
    lower_field_boxes = [None] * len(xcL)

    # Turn box corners into full box coordinate tuples
    for i in range(len(xcL)):
        upper_field_boxes[i] = (xcU[i], xcU[i] + 21, ycU[i], ycU[i] + 14)
        lower_field_boxes[i] = (xcL[i], xcL[i] + 21, ycL[i], ycL[i] + 14)

    return upper_field_boxes, lower_field_boxes


# noinspection PyTypeChecker,PyTypeChecker
def setup_for_iota_720_safe_mode2():
    # Do initializations needed for IOTA VTI timestamp extraction
    # Parameters for IOTA VTI timestamp characters when in safe mode

    # Define xy coordinates of upper field character box corners
    xcU = [78, 102, 150, 174, 222, 246, 293, 318, 341, 366, 413, 438, 462, 487]
    ycU = [200] * 14

    # Define xy coordinates of lower field character box corners
    xcL = [78, 102, 150, 174, 222, 246, 293, 318, 341, 366, 413, 438, 462, 487]
    ycL = [199] * 14

    upper_field_boxes = [None] * len(xcL)
    lower_field_boxes = [None] * len(xcL)

    # Turn box corners into full box coordinate tuples
    for i in range(len(xcL)):
        upper_field_boxes[i] = (xcU[i], xcU[i] + 21, ycU[i], ycU[i] + 14)
        lower_field_boxes[i] = (xcL[i], xcL[i] + 21, ycL[i], ycL[i] + 14)

    return upper_field_boxes, lower_field_boxes


# noinspection PyTypeChecker,PyTypeChecker
def setup_for_iota_640_safe_mode2():
    # Do initializations needed for IOTA VTI timestamp extraction
    # Parameters for IOTA VTI timestamp characters when in safe mode

    # Define xy coordinates of upper field character box corners
    xcU = [63, 85, 129, 151, 194, 216, 260, 282, 304, 326, 369, 390, 411, 434]
    ycU = [199] * 14

    # Define xy coordinates of lower field character box corners
    xcL = [63, 85, 129, 151, 194, 216, 260, 282, 304, 326, 369, 390, 411, 434]
    ycL = [199] * 14

    upper_field_boxes = [None] * len(xcL)
    lower_field_boxes = [None] * len(xcL)

    # Turn box corners into full box coordinate tuples
    for i in range(len(xcL)):
        upper_field_boxes[i] = (xcU[i], xcU[i] + 21, ycU[i], ycU[i] + 14)
        lower_field_boxes[i] = (xcL[i], xcL[i] + 21, ycL[i], ycL[i] + 14)

    return upper_field_boxes, lower_field_boxes


# noinspection PyTypeChecker,PyTypeChecker
def setup_for_iota_720_full_screen_mode3():
    # Do initializations needed for IOTA VTI timestamp extraction
    # Parameters for IOTA VTI timestamp characters when in full screen mode

    # Define xy coordinates of upper field character box corners
    xcU = [72, 96, 146, 170, 220, 244, 293, 318, 343, 367, 416, 441, 465, 490]
    ycU = [199+18] * 14

    # Define xy coordinates of lower field character box corners
    xcL = [72, 96, 146, 170, 220, 244, 293, 318, 343, 367,416, 441, 465, 490]
    ycL = [199+18] * 14

    upper_field_boxes = [None] * len(xcL)
    lower_field_boxes = [None] * len(xcL)

    # Turn box corners into full box coordinate tuples
    for i in range(len(xcL)):
        upper_field_boxes[i] = (xcU[i], xcU[i] + 21, ycU[i], ycU[i] + 14)
        lower_field_boxes[i] = (xcL[i], xcL[i] + 21, ycL[i], ycL[i] + 14)

    return upper_field_boxes, lower_field_boxes


# noinspection PyTypeChecker
def setup_for_iota_640_full_screen_mode3():
    # Do initializations needed for IOTA VTI timestamp extraction
    # Parameters for IOTA VTI timestamp characters when in full screen mode

    # Define xy coordinates of upper field character box corners
    xcU = [62, 85, 129, 151, 193, 216, 259, 281, 303, 325, 368, 390, 412, 434]
    ycU = [217] * 14

    # Define xy coordinates of lower field character box corners
    xcL = [62, 85, 129, 151, 193, 216, 259, 281, 303, 325, 368, 390, 412, 434]
    ycL = [217] * 14

    upper_field_boxes = [None] * len(xcL)
    lower_field_boxes = [None] * len(xcL)

    # Turn box corners into full box coordinate tuples
    for i in range(len(xcL)):
        upper_field_boxes[i] = (xcU[i], xcU[i] + 21, ycU[i], ycU[i] + 14)
        lower_field_boxes[i] = (xcL[i], xcL[i] + 21, ycL[i], ycL[i] + 14)

    return upper_field_boxes, lower_field_boxes


# noinspection PyTypeChecker,PyTypeChecker
def setup_for_iota_720_full_screen_mode2():
    # Do initializations needed for IOTA VTI timestamp extraction
    # Parameters for IOTA VTI timestamp characters when in full screen mode

    # Define xy coordinates of upper field character box corners
    xcU = [78, 102, 151, 175, 223, 247, 295, 318, 343, 366, 415, 439, 462, 486]
    ycU = [218] * 14

    # Define xy coordinates of lower field character box corners
    xcL = [78, 102, 151, 175, 223, 247, 295, 318, 343, 366, 415, 439, 462, 486]
    ycL = [217] * 14

    upper_field_boxes = [None] * len(xcL)
    lower_field_boxes = [None] * len(xcL)

    # Turn box corners into full box coordinate tuples
    for i in range(len(xcL)):
        upper_field_boxes[i] = (xcU[i], xcU[i] + 21, ycU[i], ycU[i] + 14)
        lower_field_boxes[i] = (xcL[i], xcL[i] + 21, ycL[i], ycL[i] + 14)

    return upper_field_boxes, lower_field_boxes


# noinspection PyTypeChecker,PyTypeChecker
def setup_for_iota_640_full_screen_mode2():
    # Do initializations needed for IOTA VTI timestamp extraction
    # Parameters for IOTA VTI timestamp characters when in full screen mode

    # Define xy coordinates of upper field character box corners
    xcU = [62, 86, 127, 151, 191, 215, 258, 282, 303, 326, 368, 391, 413, 435]
    ycU = [199+18] * 14

    # Define xy coordinates of lower field character box corners
    xcL = [62, 86, 127, 151, 191, 215, 258, 282, 303, 326, 368, 391, 413, 435]
    ycL = [199+18] * 14

    upper_field_boxes = [None] * len(xcL)
    lower_field_boxes = [None] * len(xcL)

    # Turn box corners into full box coordinate tuples
    for i in range(len(xcL)):
        upper_field_boxes[i] = (xcU[i], xcU[i] + 21, ycU[i], ycU[i] + 14)
        lower_field_boxes[i] = (xcL[i], xcL[i] + 21, ycL[i], ycL[i] + 14)

    return upper_field_boxes, lower_field_boxes


# noinspection PyTypeChecker,PyTypeChecker
def setup_for_GHS_generic():

    x0 = 391
    y0 = 193

    # Define xy coordinates of upper field character box corners
    xcU = [x0+00, x0+20, x0+60, x0+80, x0+120, x0+140, x0+180, x0+200, x0+220]
    # xcU = [362, 386, 427, 451, 491, 515, 558, 582, 603]
    ycU = [y0] * 9

    # Define xy coordinates of lower field character box corners
    xcL = [x0+00, x0+20, x0+60, x0+80, x0+120, x0+140, x0+180, x0+200, x0+220]
    # xcL = [362, 386, 427, 451, 491, 515, 558, 582, 603]
    ycL = [y0-1] * 9

    upper_field_boxes = [None] * len(xcL)
    lower_field_boxes = [None] * len(xcL)

    # Turn box corners into full box coordinate tuples
    for i in range(len(xcL)):
        upper_field_boxes[i] = (xcU[i], xcU[i] + 20, ycU[i], ycU[i] + 16)
        lower_field_boxes[i] = (xcL[i], xcL[i] + 20, ycL[i], ycL[i] + 16)

    return upper_field_boxes, lower_field_boxes


def kiwi_720_boxes(dx):
    # Define xy coordinates of upper field character box corners
    xcU = [59, 83, 130, 154, 200, 224, 295, 320, 345, 414, 438, 462]
    ycU = [205] * 12

    # Define xy coordinates of lower field character box corners
    xcL = [60, 84, 130, 154, 201, 225, 295, 320, 343, 414, 438, 462]
    ycL = [204] * 12

    xcU = [x + dx for x in xcU]
    xcL = [x + dx for x in xcL]

    return xcU, ycU, xcL, ycL

def kiwi_PAL_720_boxes(dx):
    # Define xy coordinates of upper field character box corners
    xcU = [47, 71, 118, 142, 189, 213, 283, 307, 331, 402, 426, 449]
    ycU = [253] * 12

    # Define xy coordinates of lower field character box corners
    xcL = [47, 71, 118, 142, 189, 213, 283, 307, 331, 402, 426, 449]

    ycL = [253] * 12

    xcU = [x + dx for x in xcU]
    xcL = [x + dx for x in xcL]

    return xcU, ycU, xcL, ycL


# noinspection DuplicatedCode,PyTypeChecker,PyTypeChecker
def setup_for_kiwi_PAL_720_left():

    xcU, ycU, xcL, ycL = kiwi_PAL_720_boxes(0)

    # Turn box corners into full box coordinate tuples
    upper_field_boxes = [None] * len(xcL)
    lower_field_boxes = [None] * len(xcL)
    for i in range(len(xcL)):
        upper_field_boxes[i] = (xcU[i], xcU[i] + 23, ycU[i], ycU[i] + 13)
        lower_field_boxes[i] = (xcL[i], xcL[i] + 23, ycL[i], ycL[i] + 13)

    return upper_field_boxes, lower_field_boxes


# noinspection DuplicatedCode,PyTypeChecker,PyTypeChecker
def setup_for_kiwi_PAL_720_right():

    xcU, ycU, xcL, ycL = kiwi_PAL_720_boxes(KiwiJumpAtMinuteChange)

    # Turn box corners into full box coordinate tuples
    upper_field_boxes = [None] * len(xcL)
    lower_field_boxes = [None] * len(xcL)
    for i in range(len(xcL)):
        upper_field_boxes[i] = (xcU[i], xcU[i] + 23, ycU[i], ycU[i] + 13)
        lower_field_boxes[i] = (xcL[i], xcL[i] + 23, ycL[i], ycL[i] + 13)

    return upper_field_boxes, lower_field_boxes


# noinspection DuplicatedCode,PyTypeChecker,PyTypeChecker
def setup_for_kiwi_vti_720_left():

    xcU, ycU, xcL, ycL = kiwi_720_boxes(0)

    # Turn box corners into full box coordinate tuples
    upper_field_boxes = [None] * len(xcL)
    lower_field_boxes = [None] * len(xcL)
    for i in range(len(xcL)):
        upper_field_boxes[i] = (xcU[i], xcU[i] + 23, ycU[i], ycU[i] + 13)
        lower_field_boxes[i] = (xcL[i], xcL[i] + 23, ycL[i], ycL[i] + 13)

    return upper_field_boxes, lower_field_boxes


# noinspection DuplicatedCode,PyTypeChecker,PyTypeChecker
def setup_for_kiwi_vti_720_right():

    xcU, ycU, xcL, ycL = kiwi_720_boxes(KiwiJumpAtMinuteChange)

    # Turn box corners into full box coordinate tuples
    upper_field_boxes = [None] * len(xcL)
    lower_field_boxes = [None] * len(xcL)
    for i in range(len(xcL)):
        upper_field_boxes[i] = (xcU[i], xcU[i] + 23, ycU[i], ycU[i] + 13)
        lower_field_boxes[i] = (xcL[i], xcL[i] + 23, ycL[i], ycL[i] + 13)

    return upper_field_boxes, lower_field_boxes


def kiwi_640_boxes(dx):
    # Define xy coordinates of upper field character box corners
    xcU = [49, 72, 112, 135, 177, 197, 261, 282, 302, 365, 388, 408]
    ycU = [205] * 12

    # Define xy coordinates of lower field character box corners
    xcL = [50, 73, 113, 136, 178, 198, 261, 282, 303, 365, 388, 408]
    ycL = [204] * 12

    xcU = [x + dx for x in xcU]
    xcL = [x + dx for x in xcL]

    return xcU, ycU, xcL, ycL


# noinspection DuplicatedCode,PyTypeChecker,PyTypeChecker
def setup_for_kiwi_vti_640_left():
    xcU, ycU, xcL, ycL = kiwi_640_boxes(0)

    # Turn box corners into full box coordinate tuples
    upper_field_boxes = [None] * len(xcL)
    lower_field_boxes = [None] * len(xcL)
    for i in range(len(xcL)):
        upper_field_boxes[i] = (xcU[i], xcU[i] + 23, ycU[i], ycU[i] + 13)
        lower_field_boxes[i] = (xcL[i], xcL[i] + 23, ycL[i], ycL[i] + 13)

    return upper_field_boxes, lower_field_boxes


# noinspection DuplicatedCode,PyTypeChecker,PyTypeChecker
def setup_for_kiwi_vti_640_right():
    xcU, ycU, xcL, ycL = kiwi_640_boxes(KiwiJumpAtMinuteChange)

    # Turn box corners into full box coordinate tuples
    upper_field_boxes = [None] * len(xcL)
    lower_field_boxes = [None] * len(xcL)
    for i in range(len(xcL)):
        upper_field_boxes[i] = (xcU[i], xcU[i] + 23, ycU[i], ycU[i] + 13)
        lower_field_boxes[i] = (xcL[i], xcL[i] + 23, ycL[i], ycL[i] + 13)

    return upper_field_boxes, lower_field_boxes


# noinspection PyTypeChecker,PyTypeChecker
def setup_for_boxsprite3_640():
    # Parameters for BOXSPRITE3 VTI timestamp characters width = 640

    # Define xy coordinates of lower field character box corners
    xcL = [122, 132, 150, 160, 178, 188, 205, 215, 225, 235, 253]
    ycL = [210] * len(xcL)

    # Define xy coordinates of upper field character box corners
    xcU = [122, 132, 150, 160, 178, 188, 205, 215, 225, 235, 253]
    ycU = [211] * len(xcL)

    # Turn box corners into full box coordinate tuples
    upper_field_boxes = [None] * len(xcL)
    lower_field_boxes = [None] * len(xcL)

    for i in range(len(xcL)):
        upper_field_boxes[i] = (xcU[i], xcU[i] + 9, ycU[i], ycU[i] + 14)
        lower_field_boxes[i] = (xcL[i], xcL[i] + 9, ycL[i], ycL[i] + 14)

    return upper_field_boxes, lower_field_boxes


# noinspection PyTypeChecker,PyTypeChecker
def setup_for_boxsprite3_720():
    # Parameters for BOXSPRITE3 VTI timestamp characters width = 720

    # Define xy coordinates of lower field character box corners
    xcL = [180, 192, 216, 228, 252, 264, 288, 300, 312, 324, 348]
    ycL = [202] * len(xcL)

    # Define xy coordinates of upper field character box corners
    xcU = [180, 192, 216, 228, 252, 264, 288, 300, 312, 324, 348]
    ycU = [202] * len(xcL)

    # Turn box corners into full box coordinate tuples
    upper_field_boxes = [None] * len(xcL)
    lower_field_boxes = [None] * len(xcL)

    for i in range(len(xcL)):
        upper_field_boxes[i] = (xcU[i], xcU[i] + 10, ycU[i], ycU[i] + 14)
        lower_field_boxes[i] = (xcL[i], xcL[i] + 10, ycL[i], ycL[i] + 14)

    return upper_field_boxes, lower_field_boxes


method = eval('cv2.TM_CCOEFF_NORMED')


def cv2_score(image, field_digits):
    # img = cv2.copyMakeBorder(image, 2,2,2,2, cv2.BORDER_CONSTANT, value=0)
    img = cv2.copyMakeBorder(image, 1,1,3,3, cv2.BORDER_REPLICATE, value=0)
    max_found = 0.0
    ans = 0
    max_vals = [None] * 10
    for i in range(10):
        # Apply template Matching matchTemplate needs uin8 or float32
        # When we were only dealing with uint8 from avi files, we didn't need the type conversion in the cv2 call,.
        # but aav files are uint16 (needed because they integrate 8 bit files), so we all pay the price
        res = cv2.matchTemplate(img.astype(np.float32), field_digits[i].astype(np.float32), method)
        min_val, max_val, min_loc, max_loc = cv2.minMaxLoc(res)
        max_vals[i] = max_val
        if max_val > max_found:
            max_found = max_val
            ans = i
    return ans, max_found, max_vals


def extract_timestamp(field, field_boxes, field_digits, formatter, thresh, kiwi=False, slant=False, t2fromleft=None):
    ts = ''  # ts 'means' timestamp
    _ = thresh # unused parameter

    blankscore = .5
    corr_scores = []
    digit = []
    boxsums = []
    scores = ''

    for k in range(len(field_boxes)):
        t_img = timestamp_box_image(field, field_boxes[k], kiwi, slant)
        if kiwi:  # We use boxsums to better detect empty boxes
            boxsums.append(np.sum(t_img))
            # b_img = cv2.GaussianBlur(t_img, ksize=(5, 5), sigmaX=0)
            # ans, score, _ = cv2_score(b_img, field_digits)
            ans, score, _ = cv2_score(t_img, field_digits)

            if ans == 6 or ans == 8:
                # Pick a group of test pixels at the right-hand edge (to minimize box position criticality)
                a1 = int(t_img[2, 15])  # test pixel1
                a2 = int(t_img[2, 16])  # test pixel2
                a3 = int(t_img[2, 17])  # test pixel3
                a4 = int(t_img[2, 18])  # test pixel4
                a5 = int(t_img[2, 19])  # test pixel5
                a6 = int(t_img[2, 20])  # test pixel6
                a7 = int(t_img[2, 21])  # test pixel7
                a = max(a1, a2, a3, a4, a5, a6, a7)
                b = int(t_img[3, 12])  # reference 'bright'
                c = int(t_img[2, 12])  # reference 'dark'
                ab = abs(a - b)
                ac = abs(a - c)
                if ab < ac:
                    ans = 8
                else:
                    ans = 6

            if ans == 8 or ans == 9:
                # Pick a group of test pixels at the left-hand edge (to minimize box position criticality)
                a1 = int(t_img[5, 2])  # test pixel1
                a2 = int(t_img[5, 3])  # test pixel2
                a3 = int(t_img[5, 4])  # test pixel3
                a4 = int(t_img[5, 5])  # test pixel4
                a5 = int(t_img[5, 6])  # test pixel5
                a6 = int(t_img[5, 7])  # test pixel6
                a7 = int(t_img[5, 1])  # test pixel7
                a = max(a1, a2, a3, a4, a5, a6, a7)
                b = int(t_img[6, 10])  # reference 'bright'
                c = int(t_img[5, 10])  # reference 'dark'
                ab = abs(a - b)
                ac = abs(a - c)
                if ab < ac:
                    ans = 8
                else:
                    ans = 9

        else:
            ans, score, _ = cv2_score(t_img, field_digits)

        digit.append(ans)
        corr_scores.append(score)

    # KIWI timestamp character can be blank.  We detect that as a low correlation score
    if kiwi:
        # max_sum = np.max(boxsums)
        # # text = ''
        # for i, sum in enumerate(boxsums):
        #     pix_ratio = sum / max_sum
        #     # text += f'{pix_ratio:0.2f} '
        #     if pix_ratio < 0.5:
        #         digit[i] = 0
        #         ts += ' '
        #     else:
        #         ts += f'{digit[i]}'
        # # print(text)
        for i, score in enumerate(corr_scores):
            if score < blankscore + 0.1:
                digit[i] = 0
                ts += ' '
            else:
                ts += f'{digit[i]}'
    else:
        # IOTA can also have empty selection boxes
        for i, score in enumerate(corr_scores):
            if score < blankscore:
                ts += ' '
            else:
                ts += f'{digit[i]}'

    cum_score = np.sum(corr_scores)
    for i in range(len(digit)):
        intscore = int(corr_scores[i] * 100)
        if intscore > 99:
            intscore = 99
        scores += f'{intscore:02d} '

    intcumscore = int(cum_score * 100)
    scores += f'sum: {intcumscore}'

    timestamp, time, ff_from_left = formatter(ts, t2fromleft)

    return timestamp, time, ts, scores, intcumscore, ff_from_left


def format_iota_timestamp(ts, t2fromleft):
    assert (len(ts) == 14), "len(ts) not equal to 14 in iota timestamp formatter"

    _ = t2fromleft # unused parameter

    try:
        hh = 10 * int(ts[0]) + int(ts[1])
        mm = 10 * int(ts[2]) + int(ts[3])
        ss = 10 * int(ts[4]) + int(ts[5])

        # A little hack to allow timestamp reading of Christian Weber's VTI (His VTI looks like
        # IOTA with least signicant milliseconds digit missing (blank).  We fudge in a trailing zero
        ch_list = list(ts)
        if ch_list[9] == ' ' and not(ch_list[8] == ' '):
            ch_list[9] = '0'
        if ch_list[13] == ' ' and not(ch_list[12] == ' '):
            ch_list[13] = '0'
        ts = "".join(ch_list)

        if not (ts[6] == ' ' or ts[7] == ' ' or ts[8] == ' ' or ts[9] == ' '):
            ff = 1000 * int(ts[6]) + 100 * int(ts[7]) + 10 * int(ts[8]) + int(ts[9])
        else:
            ff = 1000 * int(ts[10]) + 100 * int(ts[11]) + 10 * int(ts[12]) + int(ts[13])

        time = 3600 * hh + 60 * mm + ss + ff / 10000
        # return f'[{ts[0]}{ts[1]}:{ts[2]}{ts[3]}:{ts[4]}{ts[5]}.{ts[6]}{ts[7]}{ts[8]}{ts[9]}]', time
        return f'[{ts[0]}{ts[1]}:{ts[2]}{ts[3]}:{ts[4]}{ts[5]}.{ff:04d}]', time, None
    except ValueError:
        return f'[00:00:00.0000]', -1.0, None  # Indicate invalid timestamp by returning negative time


def format_kiwi_timestamp(ts_str, t2fromleft):

    def increment_time(hh_param, mm_param, ss_param):
        ss_param += 1
        if ss_param == 60:
            ss_param = 0
            mm_param += 1
            if mm_param == 60:
                mm_param = 0
                hh_param += 1
                if hh_param == 24:
                    hh_param = mm_param = ss_param = 0
        return hh_param, mm_param, ss_param

    assert (len(ts_str) == 12), "len(ts_str) not equal to 12 in kiwi timestamp formatter"
    ts = [0] * 12
    try:
        for i, value in enumerate(ts_str):
            # Convert spaces to zero; proper integer for rest
            if value == ' ':
                ts[i] = 0
            else:
                ts[i] = int(value)

        hh = 10 * ts[0] + ts[1]
        mm = 10 * ts[2] + ts[3]
        ss = 10 * ts[4] + ts[5]

        ff_left = 100 * ts[6] + 10 * ts[7] + ts[8]
        ff_right = 100 * ts[9] + 10 * ts[10] + ts[11]

        seconds_rolled_over = (ff_left < 17 and ff_right > 300) or (ff_right < 17 and ff_left > 300)
        if seconds_rolled_over:
            use_ff_left  = (ff_left  < 17 and ff_right > 300)
            # use_ff_right = (ff_right < 17 and ff_left  > 300)
        else:
            use_ff_left  = ff_left  > ff_right
            # use_ff_right = ff_right > ff_left

        if ff_left == 0 and not ts_str[6] == ' ':
            use_ff_left = False
            # use_ff_right = True

        if ff_right == 0 and not ts_str[9] == ' ' and seconds_rolled_over:
            use_ff_left = False
            # use_ff_right = True
            hh, mm, ss = increment_time(hh, mm, ss)
        elif ff_right == 0 and not ts_str[9] == ' ' and not seconds_rolled_over:
            use_ff_left = True
            # use_ff_right = False

        if t2fromleft is not None:
            use_ff_left = t2fromleft
            # use_ff_right = not t2fromleft

        if use_ff_left:
            ff = ff_left
            time = 3600 * hh + 60 * mm + ss + ff / 1000
            return f'[{hh:02d}:{mm:02d}:{ss:02d}.{ts[6]}{ts[7]}{ts[8]}]', time, use_ff_left
        else:
            ff = ff_right
            time = 3600 * hh + 60 * mm + ss + ff / 1000
            return f'[{hh:02d}:{mm:02d}:{ss:02d}.{ts[9]}{ts[10]}{ts[11]}]', time, use_ff_left


    except ValueError:
        return f'[00:00:00.000]', -1.0, None  # Indicate invalid timestamp by returning negative time


def format_boxsprite3_timestamp(ts, t2fromleft):
    assert (len(ts) == 11), "len(ts) not equal to 11 in boxsprite timestamp formatter"
    _ = t2fromleft # unused parameter
    try:
        hh = 10 * int(ts[0]) + int(ts[1])
        mm = 10 * int(ts[2]) + int(ts[3])
        ss = 10 * int(ts[4]) + int(ts[5])
        ff = 1000 * int(ts[7]) + 100 * int(ts[8]) + 10 * int(ts[9]) + int(ts[10])
        time = 3600 * hh + 60 * mm + ss + ff / 10000
        return f'[{ts[0]}{ts[1]}:{ts[2]}{ts[3]}:{ts[4]}{ts[5]}.{ts[7]}{ts[8]}{ts[9]}{ts[10]}]', time, None
    except ValueError:
        return f'[00:00:00.0000]', -1.0, None  # Indicate invalid timestamp by returning negative time

def format_ghs_timestamp(ts, t2fromleft):
    assert (len(ts) == 9), "len(ts) not equal to 9 in GHS timestamp formatter"
    _ = t2fromleft # unused parameter
    try:
        hh = 10 * int(ts[0]) + int(ts[1])
        mm = 10 * int(ts[2]) + int(ts[3])
        ss = 10 * int(ts[4]) + int(ts[5])
        ff = 100 * int(ts[6]) + 10 * int(ts[7]) + int(ts[8])
        time = 3600 * hh + 60 * mm + ss + ff / 1000
        return f'[{ts[0]}{ts[1]}:{ts[2]}{ts[3]}:{ts[4]}{ts[5]}.{ts[6]}{ts[7]}{ts[8]}]', time, None
    except ValueError:
        return f'[00:00:00.000]', -1.0, None  # Indicate invalid timestamp by returning negative time

def timestamp_box_image(img, box, kiwi, slant):
    # Note: img must be in field mode
    (xL, xR, yL, yU) = box
    if not (kiwi or slant):
        return img[yL:yU+1, xL:xR+1].copy()
    elif slant:
        w = xR - xL + 1
        h = yU - yL + 1
        rh = int(h/2) + 1
        patch = np.ndarray((rh, w), dtype='uint8')

        offset = 6
        img_row = yL
        patch[0, :] = img[img_row, (xL + offset):(xR + offset + 1)]

        offset = 4
        img_row += 2
        patch[1, :] = img[img_row, (xL + offset):(xR + offset + 1)]

        offset = 2
        img_row += 2
        patch[2, :] = img[img_row, (xL + offset):(xR + offset + 1)]

        offset = 0
        img_row += 2
        patch[3, :] = img[img_row, (xL + offset):(xR + offset + 1)]

        offset = -1
        img_row += 2
        patch[4, :] = img[img_row, (xL + offset):(xR + offset + 1)]

        offset = -3
        img_row += 2
        patch[5, :] = img[img_row, (xL + offset):(xR + offset + 1)]

        offset = -5
        img_row += 2
        patch[6, :] = img[img_row, (xL + offset):(xR + offset + 1)]


        return patch

    else:
        w = xR - xL + 1
        h = yU - yL + 1
        rh = int(h / 2) + 1
        patch = np.ndarray((rh, w), dtype='uint8')

        offset = 0
        img_row = yL
        patch[0, :] = img[img_row, (xL + offset):(xR + offset + 1)]

        offset = 0
        img_row += 2
        patch[1, :] = img[img_row, (xL + offset):(xR + offset + 1)]

        offset = 0
        img_row += 2
        patch[2, :] = img[img_row, (xL + offset):(xR + offset + 1)]

        offset = 0
        img_row += 2
        patch[3, :] = img[img_row, (xL + offset):(xR + offset + 1)]

        offset = 0
        img_row += 2
        patch[4, :] = img[img_row, (xL + offset):(xR + offset + 1)]

        offset = 0
        img_row += 2
        patch[5, :] = img[img_row, (xL + offset):(xR + offset + 1)]

        offset = 0
        img_row += 2
        patch[6, :] = img[img_row, (xL + offset):(xR + offset + 1)]

        return patch


def print_confusion_matrix(field_digits, printer):
    # Compute a 'confusion matrix' which compares the match coefficient
    # of each sample digit against all the other sample digits.

    c = np.ndarray((10,10))
    for sample in range(10):
        img = field_digits[sample].copy()
        #imgb = cv2.GaussianBlur(img.copy(), (3,3), 0)
        ans = cv2_score(img, field_digits)
        c[sample,:] = ans[2]

    # Pretty print the matrix
    printer(msg=f'Confusion matrix (scores sample digits against sample digits)', blankLine=False)
    printer(msg=f'     0    1    2    3    4    5    6    7    8    9', blankLine=False)
    for i in range(10):
        line_format = f'{i} '
        line_format += f'{c[i,0]:5.2f}{c[i,1]:5.2f}'
        line_format += f'{c[i,2]:5.2f}{c[i,3]:5.2f}'
        line_format += f'{c[i,4]:5.2f}{c[i,5]:5.2f}'
        line_format += f'{c[i,6]:5.2f}{c[i,7]:5.2f}'
        line_format += f'{c[i,8]:5.2f}{c[i,9]:5.2f}'
        printer(msg=line_format, blankLine=False)


def locate_timestamp_vertically(img, fig, showplot=False):
    vert_profile = []
    for i in range(img.shape[0]):
        vert_profile.append(np.mean(img[i, :]))

    med_val = np.median(vert_profile)
    max_val = np.max(vert_profile)
    thresh = (max_val - med_val) / 2 + med_val

    top = None
    bottom = None

    for i in range(len(vert_profile)):
        if vert_profile[i] >= thresh:
            top = i
            break

    for i in range(len(vert_profile) - 1, 0, -1):
        if vert_profile[i] >= thresh:
            bottom = i
            break

    if showplot:
        plt.figure(fig, figsize=(10, 6))
        plt.plot(vert_profile, 'bo')
        plt.xlabel('y (row) coordinate within field')
        plt.ylabel('Pixel averages across row')
        plt.title('The red lines show where, on the y axis, the timestamp characters are located')
        plt.vlines([top, bottom], ymin=np.min(vert_profile), ymax=np.max(vert_profile), color='r')
        plt.grid()
        plt.show()

    return top, bottom



import pyqtgraph as pg
from PyQt5.QtCore import QRectF


class OcrAperture(pg.GraphicsObject):

    # Initialize aperture specified by a field box
    def __init__(self, fieldbox, boxnum, position, kiwi):
        self.kiwiStyle = kiwi
        self.boxnum = boxnum
        self.joggable = False
        self.position = position  # 'upper'  or 'lower'
        self.pen = pg.mkPen('r')
        self.color = 'red'
        xL, xR, yU, yL = fieldbox
        self.x0 = xL
        self.y0 = yU
        self.xsize = xR - xL
        self.ysize = yL - yU
        self.menu = None

        # note that the use of super() is often avoided because Qt does not
        # allow to inherit from multiple QObject subclasses.
        pg.GraphicsObject.__init__(self)

    def setBox(self, fieldbox):
        xL, xR, yU, yL = fieldbox
        self.x0 = xL
        self.y0 = yU
        self.xsize = xR - xL
        self.ysize = yL - yU

    def getBox(self):  # Return 'internal' box
        xR = self.xsize + self.x0
        yL = self.ysize + self.y0
        return self.x0, xR - 1, self.y0, yL - 1  # (xL, xR, yU, yL)

    # All graphics items must have boundingRect() defined.
    def boundingRect(self):
        return QRectF(self.x0, self.y0, self.xsize, self.ysize)

    # All graphics items must have paint() defined.
    def paint(self, p, *args):
        _ = args # unused parameter
        p.setPen(self.pen)
        if self.joggable:
            p.setPen(pg.mkPen('y'))
        else:
            p.setPen(pg.mkPen('r'))
        if not self.kiwiStyle:
            p.drawRect(self.boundingRect())
        else:
            # p.setPen(pg.mkPen('y'))
            # p.setPen(pg.mkPen('r'))
            dx = 5
            x0 = self.x0 + dx
            y0 = self.y0
            x1 = x0 + self.xsize
            y1 = y0
            x2 = self.x0 - dx
            y2 = self.y0 + self.ysize
            x3 = x2 + self.xsize
            y3 = y2
            p.drawLine(x0, y0, x1, y1) # Top line
            p.drawLine(x0, y0, x2, y2) # Left line
            p.drawLine(x1, y1, x3, y3) # Right line
            p.drawLine(x2, y2, x3, y3) # Bottom line



# -*- coding: utf-8 -*-

# Form implementation generated from reading ui file 'hotPixelDialog.ui'
#
# Created by: PyQt5 UI code generator 5.12.1
#
# WARNING! All changes made in this file will be lost!

from PyQt5 import QtCore, QtGui, QtWidgets


class Ui_hotPixelThresholdDialog(object):
    def setupUi(self, hotPixelThresholdDialog):
        hotPixelThresholdDialog.setObjectName("hotPixelThresholdDialog")
        hotPixelThresholdDialog.setWindowModality(QtCore.Qt.NonModal)
        hotPixelThresholdDialog.resize(700, 212)
        sizePolicy = QtWidgets.QSizePolicy(QtWidgets.QSizePolicy.Fixed, QtWidgets.QSizePolicy.Fixed)
        sizePolicy.setHorizontalStretch(0)
        sizePolicy.setVerticalStretch(0)
        sizePolicy.setHeightForWidth(hotPixelThresholdDialog.sizePolicy().hasHeightForWidth())
        hotPixelThresholdDialog.setSizePolicy(sizePolicy)
        hotPixelThresholdDialog.setMinimumSize(QtCore.QSize(700, 212))
        hotPixelThresholdDialog.setMaximumSize(QtCore.QSize(700, 212))
        self.gridLayout = QtWidgets.QGridLayout(hotPixelThresholdDialog)
        self.gridLayout.setObjectName("gridLayout")
        self.verticalLayout = QtWidgets.QVBoxLayout()
        self.verticalLayout.setObjectName("verticalLayout")
        self.plainTextEdit = QtWidgets.QPlainTextEdit(hotPixelThresholdDialog)
        self.plainTextEdit.setReadOnly(True)
        self.plainTextEdit.setObjectName("plainTextEdit")
        self.verticalLayout.addWidget(self.plainTextEdit)
        self.horizontalLayout = QtWidgets.QHBoxLayout()
        self.horizontalLayout.setObjectName("horizontalLayout")
        spacerItem = QtWidgets.QSpacerItem(40, 20, QtWidgets.QSizePolicy.Expanding, QtWidgets.QSizePolicy.Minimum)
        self.horizontalLayout.addItem(spacerItem)
        self.hotPixelThresholdEdit = QtWidgets.QLineEdit(hotPixelThresholdDialog)
        self.hotPixelThresholdEdit.setMaximumSize(QtCore.QSize(80, 16777215))
        self.hotPixelThresholdEdit.setInputMethodHints(QtCore.Qt.ImhPreferNumbers)
        self.hotPixelThresholdEdit.setMaxLength(6)
        self.hotPixelThresholdEdit.setObjectName("hotPixelThresholdEdit")
        self.horizontalLayout.addWidget(self.hotPixelThresholdEdit)
        self.label = QtWidgets.QLabel(hotPixelThresholdDialog)
        self.label.setObjectName("label")
        self.horizontalLayout.addWidget(self.label)
        spacerItem1 = QtWidgets.QSpacerItem(40, 20, QtWidgets.QSizePolicy.Expanding, QtWidgets.QSizePolicy.Minimum)
        self.horizontalLayout.addItem(spacerItem1)
        self.verticalLayout.addLayout(self.horizontalLayout)
        self.gridLayout.addLayout(self.verticalLayout, 0, 0, 1, 1)
        self.buttonBox = QtWidgets.QDialogButtonBox(hotPixelThresholdDialog)
        self.buttonBox.setOrientation(QtCore.Qt.Vertical)
        self.buttonBox.setStandardButtons(QtWidgets.QDialogButtonBox.Cancel|QtWidgets.QDialogButtonBox.Ok)
        self.buttonBox.setObjectName("buttonBox")
        self.gridLayout.addWidget(self.buttonBox, 0, 1, 1, 1)

        self.retranslateUi(hotPixelThresholdDialog)
        self.buttonBox.accepted.connect(hotPixelThresholdDialog.accept)
        self.buttonBox.rejected.connect(hotPixelThresholdDialog.reject)
        QtCore.QMetaObject.connectSlotsByName(hotPixelThresholdDialog)

    def retranslateUi(self, hotPixelThresholdDialog):
        _translate = QtCore.QCoreApplication.translate
        hotPixelThresholdDialog.setWindowTitle(_translate("hotPixelThresholdDialog", "Hot pixel threshold dialog"))
        self.plainTextEdit.setPlainText(_translate("hotPixelThresholdDialog", "Unfortunately, it is necessary to set the threshold for hot-pixels manually. This is because\n"
"automatic methods fail sometimes, mainly when the background is \'clipped\' or otherwise\n"
"has very little noise.  In such cases, the simple idea of setting a threshold based on say\n"
"3 sigma above the mean fails because sigma will be 0.\n"
"\n"
"A good way to see what value to set for threshold is to use the Plot Robust Mean button\n"
"to get a nice view of the values contained in the apertures that you placed at the site of\n"
"hot-pixel groups."))
        self.label.setText(_translate("hotPixelThresholdDialog", "hot-pixel threshold"))


